/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef GPU_MEMORY_CUH_
#define GPU_MEMORY_CUH_


#define PAD(N,pad)  (N+(N % pad ? N-(N % pad) : 0))

#define MANAGE_ERROR_HOST(param) 			\
	if(param != cudaSuccess) {			\
	printf("\n\nGRAPHIC CARD ERROR : %s \nABORT... \n",cudaGetErrorString(param)); \
	exit(0); }

#include <vector>
#include "utils.hpp"



#include "integer.cuh"

/************************************************************
 *  Memory class for integers manipulation                  *
 *  data on GPU ared in coalesced form with some padding    *
 ************************************************************/


template<uint N, uint size>
class Memory {
public:
	// allocate space on the GPU 
	_H_ static void allocGPU(ushort2*&v){
		const uint pad = PAD(size,16);
		MANAGE_ERROR_HOST( cudaMalloc((void**) &v, (sizeof(ushort2)*pad*NB_LIMB>>1) ));
	}

	// free space on the GPU 
	_H_ static void freeGPU(ushort2*&v){
		cudaFree(v);
	}

	// copy vector of integer from CPU to GPU (allocation is done on GPU side)
	template <typename Alloc>
	_H_ static void sendToGPU(ushort2*& v, const vector<Integer<N, Alloc > >& vec)
	{
		const uint pad = PAD((size >> 1),16);
		ushort2 linear_padded_data[NB_LIMB*pad];
		for(uint i = 0; i < size; i++) { 
			for(uint j = 0; j < NB_LIMB-1; j = j+2) {
				linear_padded_data[(j*pad)+i] = make_ushort2(vec[i][j], vec[i][j+1]);
			}
			if(NB_LIMB & 1) linear_padded_data[((NB_LIMB-1)*pad) + i] = make_ushort2(vec[i][NB_LIMB-1],0);
		}
		MANAGE_ERROR_HOST( cudaMalloc((void**) &v, (sizeof(ushort2)*pad*NB_LIMB)));
		MANAGE_ERROR_HOST( cudaMemcpy(v,linear_padded_data, (sizeof(ushort2)*pad*NB_LIMB), cudaMemcpyHostToDevice));
	}
	
	// copy vector of uint from CPU to GPU (allocation is done on GPU side)
	_H_ static void sendToGPU(uint*& v, const vector<uint>& vec)
	{
		const uint pad = PAD(size,16);
		uint linear_padded_data[pad];
		for(uint i = 0; i < size; i++) 
			linear_padded_data[i] = vec[i];
		MANAGE_ERROR_HOST( cudaMalloc((void**) &v, sizeof(uint)*pad));
		MANAGE_ERROR_HOST( cudaMemcpy(v,linear_padded_data, sizeof(uint)*pad, cudaMemcpyHostToDevice));
	} 

	// copy vector of integer from gpu to cpu
	template <typename Alloc>
	_H_ static void sendToCPU(vector<Integer<N, Alloc> >& vec, ushort2* v)
	{
		const ushort pad = PAD((size >> 1),16);		
		ushort2 mp[pad*NB_LIMB];		
		MANAGE_ERROR_HOST( cudaMemcpy(mp,v,(sizeof(ushort2)*pad*NB_LIMB), cudaMemcpyDeviceToHost));
		for(ushort i = 0; i < size; i++) {
			for(uint j = 0; j < NB_LIMB-1; j = j+2) {
				(vec[i])[j]   = mp[(j*pad)+i].x;
				(vec[i])[j+1] = mp[(j*pad)+i].y;
			}
			if(NB_LIMB & 1) (vec[i])[NB_LIMB-1] = mp[(NB_LIMB-1)].x;
		}
	}
	
	_H_ static void sendToGPUTexture(const vector<Integer<N, LocalMemory> >& v)
	{		
		cudaArray* _array;
		cudaChannelFormatDesc _desc = cudaCreateChannelDesc<ushort>(); 
		cudaMallocArray(&_array,&_desc, 2*NB_LIMB,v.size());
		for (uint i=0;i<v.size();++i){
			cudaMemcpyToArray(_array,0,i,v[i]._alloc._mp_d ,2*NB_LIMB,cudaMemcpyHostToDevice);
		}
		
		cudaBindTextureToArray(GPUTexture, _array);//, &_desc); 
	} 

	// uncoalesce integer from global memory at position k
	template <typename Alloc>
	_D_ static void globalLoad(Integer<N, Alloc>& x, ushort2* v, ushort k)  { /* nothing to do, see specialization below*/}

	_D_ static void globalLoad(Integer<N, LocalMemory>& x, ushort2* v, uint k)  
	{
		const ushort pad = PAD((size >> 1),16); 
#pragma unroll
		for(ushort i = 0; i < NB_LIMB-1; i = i+2) {
			x[i]   = v[(i*pad)+k].x;
			x[i+1] = v[(i*pad)+k].y;
		}
		if(NB_LIMB % 1) x[NB_LIMB-1] = v[((NB_LIMB-1) * pad)+k].x;
	}

	// coalesce integer to global memory at position k
	template<typename Alloc>
	_D_ static void globalStore(ushort2* v, const Integer<N, Alloc>& x,  ushort k){ /* nothing to do, see specialization below*/}
	
	_D_ static void globalStore(ushort2* v, const Integer<N, LocalMemory>& x,  ushort k)
	{
		const int pad = PAD((size >> 1),16); 
#pragma unroll
		for(ushort i = 0; i < NB_LIMB-1; i = i+2) {
			v[(i*pad)+k].x = x[i];
			v[(i*pad)+k].y = x[i+1];
		}
		if(NB_LIMB % 1) v[((NB_LIMB-1) *  pad) + k].x = x[NB_LIMB-1];
	}
};



/************************************************************
 *  structure for data involved in the computation          * 
 ************************************************************/

template<uint N>
struct GPU_vector_data{
	
	// pointer to coalesced memory on GPU for Fp definition of the Elliptic Curve (prime and prime^(-1) mod 2^16)
	ushort2* __v_prime;
	uint*    __v_invp;

	// pointer to coalesced memory on GPU for parameters a of the Elliptic Curve
	ushort2*  __v_a;
	
	// pointer to coalesced memory on GPU for scalars k in [k]P
	ushort2* __v_k;

	// pointer to coalesced memory on GPU for coordinates (x,y,z) of P in [k]P
	ushort2* __v_Px;
	ushort2* __v_Py;
	ushort2* __v_Pz;

	// pointer to coalesced memory on GPU for coordinates (x,y,z) of [k]P
	ushort2*  __v_Qx;
	ushort2*  __v_Qy;
	ushort2*  __v_Qz;
};

#endif //GPU_MEMORY_CUH_
