/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef _WNAF_CUH_
#define _WNAF_CUH_
#include <string>
#include <cuda.h>

#include "integer.cuh"

#ifdef NAF3
#define NAF_W 3
#else
#define NAF_W 2
#endif

#define SIZE 32 
typedef uint wnafword;

// W bits (11..1) wide mask
#define WMASK ((1<<W)-1)

// Usefull macros
#define CEILDIV(x,y) (( (x-y*(x/y))>0 ? x/y+1 : x/y ))

// Nb WNAF digit per word
#define NDW (SIZE/W)

// Nb words (for a N bits value)
#define NW (CEILDIV((N+1), NDW))

// -------------------------------------------------------------------
// Basic version

template<uint N>
class Naf {
public:
	uint  _msd_rank;  // in [0,...,N]
	short _d[N+1]; 
	
	_HD_ Naf(){}

	_D_ void init(short x=0, uint nb=N+1)
	{
		this->_msd_rank = nb-1;
#pragma unroll
		for(uint i=0; i<nb; i++)
			this->_d[i] = x;
	}

	// WARNING: strong assumption x != 0
	template<typename Allocator>
	_D_ void from_integer(const Integer<N,Allocator>& x) 
	{
		Integer<N,Allocator> r(x);	

		uint i=0;
		while (r.is_gt0())
			{
				if (r.is_odd())
					{
						int d = r.mods(NAF_W); // x mods 2^W
						this->_d[i] = d;
						this->_msd_rank = i;
						r.sub_in(d);
					}
				else
					{
						this->_d[i] = 0;
					}
				r.div2();
				i++;
			}
		
		// set "unsed" MSD to 0
#pragma unroll
 		for(uint j=i-1; j<N+1; j++)
 			this->_d[j] = 0;
	}

	_D_ short get_digit_at(int pos) const { return this->_d[pos]; }

	_H_ std::string to_string() const 
	{ 
		ostringstream os;
		for (uint i=0; i<N+1; i++)
			os << this->_d[i];
		return os.str();    
	}
};

// -------------------------------------------------------------------


// N bits integer => at most N+1 NAF-digits
template<uint N, uint W>
class OWNaf
{
public:
	uint _msd_rank;
	wnafword _v[NW];
	
	_HD_ OWNaf(){}

	_D_ void init(wnafword x, uint nb=NW)
	{
#pragma unroll
		for (uint i=0; i<nb; i++)
			this->_v[i] = x;
	}
	
	_D_ short get_digit_at(uint i) const
	{
		wnafword t, s, d;
		t = this->_v[i/NDW];
		t &= WMASK << (W*(i%NDW));
		t >>= W*(i%NDW);
		s = (t & (1<<(W-1))) >> (W-1); // sign bit (0 neg., 1 pos.)
		d = t & ( (1<<(W-1))-1 ); // |digit|
		return (1-2*s)*d;
	}
	
	_D_ void set_digit_at(uint i, short x)
	{
		wnafword mz, t, d;
		t = this->_v[i/NDW];
		if (x<0)
			d = (1<<(W-1)) - x;
		else
			d = x;
		mz = -1 ^ ( ((1<<W)-1) << (W*(i%NDW)) );
		t &= mz;
		t += d << (W*(i%NDW));
		this->_v[i/NDW] = t;    
	}
	
	template<typename Allocator>
	_D_ void from_integer(const Integer<N,Allocator>& x) 
	{
		Integer<N,Allocator> r(x);	
		uint i=0;

		this->init(0);

		while (r.is_gt0())
			{
				if (r.is_odd())
					{
						int d = r.mods(NAF_W); // x mods 2^W
						this->set_digit_at(i, d);
						this->_msd_rank = i;
						r.sub_in(d);
					}
				else
					{
						this->set_digit_at(i, 0);
					}
				r.div2();
				i++;
			}
		
		// set "unsed" MSD to 0
#pragma unroll
 		for(uint j=i-1; j<N+1; j++)
			this->set_digit_at(i, 0);
	}
	
	_H_ std::string to_string() const 
	{ 
		ostringstream os;
		for (int i=this->_msd_rank; i>=0; i--)
			os << this->get_digit_at(i) << ' ';
		return os.str();    
	}
	
};

#endif // _WNAF_CUH_
